#! /usr/local/bin/perl

#----------------------------------------------------------------------------------------------
#	Author: Ravi Vijaya Satya, BHSAI
#	File: FilterSNPs.pl
#	Purpose: Take a tabulay SNP/Indel file, and filter SNPs within 
# 	certain distance of each other 
#	the description line of each sequence is expected to contain start: end: length: etc. 
#-----------------------------------------------------------------------------------------------

use strict;
use warnings;
use Bio::Seq;
use Bio::SeqIO;
use File::Temp qw/ tempfile /;

if(!@ARGV or $#ARGV < 4){
	die("Usage: perl $0 InputSNPFile MinExactMatch MinDistanceFromEnds MaskLowQuality OutSNPFile");
}

our ($InputSNPFile, $MinExactMatch, $MinDistanceFromEnds, $MaskLowQualityBases, $OutSNPFile) = @ARGV[0..4];

my $FH;
my @Lines;
my $LineCtr = 0;
my @LineStatus;

my ($NextValidRefPos, $PolyType, $RefID, $QryID, $RefPos, $QryPos, $Alleles, $RAllele, $QAllele);
my @tags;

my ($PrevRefID, $PrevQryID) = ("", "");
my ($RefStart, $RefEnd, $QryStart, $QryEnd);
my ($UCRAllele, $UCQAllele);

$NextValidRefPos = $MinExactMatch+1;

open($FH, "<$InputSNPFile");

while (<$FH>){
	$Lines[++$LineCtr] = $_;
	$LineStatus[$LineCtr] = "valid";
	if(/(\S+)\s+(\S+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\S+)/){
		($PolyType, $RefID, $QryID, $RefPos, $QryPos, $Alleles) = ($1, $2, $3, $4, $5, $6);
		($RAllele, $QAllele) = split(/\|/, $Alleles);
		
		# check if the alleles are low quality
		if($MaskLowQualityBases ==1 ){
			$UCRAllele = uc($RAllele);
			$UCQAllele = uc($QAllele);
			if( ($UCRAllele ne $RAllele) || ($UCQAllele ne $QAllele)){ 		# atleast one of them is lower case, meaning low quality
				$LineStatus[$LineCtr] = "invalid";
				next;
			}
		}
		# check if the SNP is away from the ends of both the ref and query sequences
		@tags = split(/\|/, $RefID);
		if($tags[1] =~ /start:(\d+)/){
			$RefStart = $1;
		}
		if($tags[2] =~ /end:(\d+)/){
			$RefEnd = $1;
		}
		if(($RefPos - $RefStart) < $MinDistanceFromEnds || ($RefEnd-$RefPos-length($RAllele)+1) < $MinDistanceFromEnds){
			$LineStatus[$LineCtr] = "invalid";
		} 
		
		@tags = split(/\|/, $QryID);
		if($tags[1] =~ /start:(\d+)/){
			$QryStart = $1;
		}
		if($tags[2] =~ /end:(\d+)/){
			$QryEnd = $1;
		}
		if( ($QryPos-$QryStart) < $MinDistanceFromEnds || ($QryEnd-$QryPos-length($QAllele)+1) < $MinDistanceFromEnds){
			$LineStatus[$LineCtr] = "invalid";
		}	
			
		# check if current line is allowable
		if($PrevRefID ne $RefID){
			$PrevRefID = $RefID;
			$NextValidRefPos = $RefStart+$MinExactMatch;
		}		
		if($RefPos < $NextValidRefPos){
			$LineStatus[$LineCtr] = "invalid";
			$LineStatus[$LineCtr-1] = "invalid";
			$NextValidRefPos = $RefPos + length($RAllele)+$MinExactMatch;
			next;
		}
		$NextValidRefPos = $RefPos + length($RAllele)+$MinExactMatch;

	}
}
close $FH;

open($FH, ">$OutSNPFile");

for(my $i = 1; $i <= $LineCtr; $i++){
	if($LineStatus[$i] eq "valid"){
		print $FH $Lines[$i];
	}
}

close $FH;


