#! /usr/local/bin/perl

#----------------------------------------------------------------------------------------------
#	Author: Ravi Vijaya Satya, BHSAI
#	File: FindPairWiseSNPs.pl
#	Purpose: Takes an input multifasta file, properly indexed, and a Multi Fasta query file. 
#	The output is the set of SNPs and indels between the two genomes
# 	The input file should be a candidate file that has atleast one candidate sequence 
#	the description line of each sequence is expected to contain start: end: length: etc. 
#-----------------------------------------------------------------------------------------------

use strict;
use warnings;
use Env;
use Bio::Seq;
use Bio::SeqIO;
my $MummerDir=$MUMMERDIR;

sub run_nucmer_multi($); 
sub get_cands(\$\@);
sub load_file($$);
sub build_alignment(\@\@$$$$\@\$);
sub find_indels($$$$);

my $MAX_NBR_SIZE = 100000000;			# Max 100 MB to ensure that there will be no memory issues

my $RefHash = {
	"id"		=> [],
	"seq"		=> [],
	"name"		=> [],
	"start"		=> [],
	"end"		=> [],
	"length"	=> [],
	"count"		=> 0,
	"number"	=> {},
};

my $QryHash = {
	"id"		=> [],
	"seq"		=> [],
	"name"		=> [],
	"start"		=> [],
	"end"		=> [],
	"length"	=> [],
	"count"		=> 0,
	"number"	=> {},
};


if(!@ARGV or $#ARGV < 6){
	die("Usage: perl $0 target.fasta nbr.fasta MinClusterLen MinExactMatch MaxGaps MinIndelSize out_prefix");
}

our ($TargetFileName, $QueryFileName, $MinClusterLen, $MinExactMatch, $MaxGaps, $MinIndelSize, $OutPrefix) = @ARGV[0..6];

load_file($TargetFileName, $RefHash);
load_file($QueryFileName, $QryHash);

# run nucmer and mark common sequences
&run_nucmer($TargetFileName, $QueryFileName);

find_indels($RefHash, $QryHash, $OutPrefix.".delta", $OutPrefix.".lindels");


sub run_nucmer($) {
	
	my($RefFileName, $NbrFileName) = @_;	
	my $cand_cnt = 0;
	
#	run mummer
	
	my $Nucmer_cmd = "$MummerDir/nucmer  $RefFileName   $NbrFileName -c $MinClusterLen -l $MinExactMatch -g $MaxGaps --noextend --mumreference -p $OutPrefix";
	#print "running $Nucmer_cmd\n";
	system($Nucmer_cmd);
	
	my $Filter_cmd = "$MummerDir/delta-filter -r -q $OutPrefix.delta > $OutPrefix.filtered.delta";
    print "running $Filter_cmd\n";
    system ($Filter_cmd);
    
    my $Snp_Cmd = "$MummerDir/show-snps -Clr $OutPrefix.filtered.delta > $OutPrefix.snps";
    system($Snp_Cmd);
 
 # parse output of show-snps
 
 my ($SNPOutFH, $PolyFH);
 
 open($SNPOutFH, "<$OutPrefix.snps");
 open($PolyFH, ">$OutPrefix.poly");
 
 my ($RPos, $RChar, $QPos, $QChar, $RID, $QID, $PrevRID, $PrevQID, $RNo, $QNo);
 my ($StartRPos, $StartQPos, $PrevRPos, $PrevQPos, $PolyType) = (-1,-1,-1,-1,0);		# Polytype: 0 - SNP; 1 - Insertion; -1 - Deletion;
 my $IndelSeq = "";
 while(<$SNPOutFH>){
 	if(/(\d+)\s+(\S)\s(\S)\s+(\d+)\D+(\d+)\D+(\d+)\D+(\d+)\D+(\d+)\D+(\d)\D+(\d)\s+(\S+)\s+(\S+)/){
 		($RPos, $RChar, $QChar, $QPos, $RID, $QID) = ($1, $2, $3, $4, $11, $12);
 		$RNo = $RefHash->{number}{$RID};
 		$QNo = $QryHash->{number}{$QID};
 		$RPos = $RPos + $RefHash->{start}[$RNo] -1;
 		$QPos = $QPos + $QryHash->{start}[$QNo] -1;
 		
 		if($PolyType != 0){
 			if($RChar eq "."){
 				if($PolyType == 1 && $PrevQPos == ($QPos-1)){
 					$PrevQPos = $QPos;
 					$IndelSeq .= $QChar;
 				}else{
 					if($PolyType == 1){
 						print $PolyFH "INS\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t.|".$IndelSeq."\n";
 					}else{
 						print $PolyFH "DEL\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t".$IndelSeq."|.\n";
 					}
 					($PolyType, $StartRPos, $PrevRPos, $StartQPos, $PrevQPos, $PrevRID, $PrevQID) = (1, $RPos, $RPos, $QPos, $QPos, $RID, $QID);
 					$IndelSeq = $QChar;
 				}
 			}elsif($QChar eq "."){
 				if($PolyType == -1 && $PrevRPos == ($RPos-1)){
 					$PrevRPos = $RPos;
 					$IndelSeq .= $RChar;
 				}else{
 					if($PolyType == 1){
 						print $PolyFH "INS\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t.|".$IndelSeq."\n";
 					}else{
 						print $PolyFH "DEL\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t".$IndelSeq."|.\n";
 					}
 					($PolyType, $StartRPos, $PrevRPos, $StartQPos, $PrevQPos, $PrevRID, $PrevQID) = (-1, $RPos, $RPos, $QPos, $QPos, $RID, $QID);
 					$IndelSeq = $RChar;
 				}
 			}else{
 				if($PolyType == 1){
 						print $PolyFH "INS\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t.|".$IndelSeq."\n";
 				}else{
 						print $PolyFH "DEL\t".$PrevRID."\t".$PrevQID."\t".$StartRPos."\t".$StartQPos."\t".$IndelSeq."|.\n";
 				}
 				print $PolyFH "SNP\t".$RID."\t".$QID."\t".$RPos."\t".$QPos."\t".$RChar."|".$QChar."\n";
 				$PolyType = 0;
 				$IndelSeq = "";
 			}
 		}else{
 			if($RChar eq "."){
 				($PolyType, $StartRPos, $PrevRPos, $StartQPos, $PrevQPos, $PrevRID, $PrevQID) = (1, $RPos, $RPos, $QPos, $QPos, $RID, $QID);
 				$IndelSeq = $QChar;
 			}elsif($QChar eq "."){
 				($PolyType, $StartRPos, $PrevRPos, $StartQPos, $PrevQPos, $PrevRID, $PrevQID) = (-1, $RPos, $RPos, $QPos, $QPos, $RID, $QID);
 				$IndelSeq = $RChar;
 			}else{
 				print $PolyFH "SNP\t".$RID."\t".$QID."\t".$RPos."\t".$QPos."\t".$RChar."|".$QChar."\n";
 				$IndelSeq = "";
 			} 			
 		}
 	}
 }

 
 close($SNPOutFH);
 close($PolyFH);
}
 
 sub load_file($$){
 	my ($FileName, $HashRef) = @_;
 	my $SeqFile;
 	
	eval{ 
		$SeqFile = new Bio::SeqIO(-format => 'fasta', -file => $FileName);
	};

	if($@){
		die("Problem opening fasta format file $FileName", $@);
	}

	# read the input sequences
	my $Next_Cand;
	my $CandCnt = 0;
	while($Next_Cand = $SeqFile->next_seq()){
		$HashRef->{seq}[$CandCnt] = $Next_Cand->seq();
		if($Next_Cand->id() =~ /^(\w+)\D+(\d+)\D+(\d+)\D+(\d+)/){
			#print "parsing candidate $CandTags[1]; $1, $2, $3, $4\n";
			$HashRef->{id}[$CandCnt] 		= $Next_Cand->id();
			$HashRef->{name}[$CandCnt] 		= $1;
			$HashRef->{start}[$CandCnt] 	= $2;
			$HashRef->{end}[$CandCnt]		= $3;
			$HashRef->{len}[$CandCnt] 		= $4;
			$HashRef->{number}{$HashRef->{id}[$CandCnt]}			= $CandCnt;
		}else{
			die("incorrect format in the input candidate file \n");
		}
		#print "Candidate $CandCnt, ".$HashRef->{id}[$CandCnt].", ".$HashRef->{name}[$CandCnt].", ".$HashRef->{start}[$CandCnt].", ".$HashRef->{end}[$CandCnt].", ".$HashRef->{len}[$CandCnt]."\n";
		$CandCnt++;
		$HashRef->{count}	= $CandCnt;
	}
}

sub find_indels($$$$){
	my ($RHRef, $QHRef, $DeltaFileName, $IndelFileName) = @_;
	
	my ($DeltaFH, $IndelFH);
	
	open($DeltaFH, "<$DeltaFileName");
	open($IndelFH, ">$IndelFileName");
	
	my (@RefSeqs, @QrySeqs);
	for(my $i = 0; $i < $RHRef->{count}; $i++){
		$RefSeqs[$i] = [split(//, $RHRef->{seq}[$i])];
	}
	
	for(my $i = 0; $i < $QHRef->{count}; $i++){
		$QrySeqs[$i] = [split(//, $QHRef->{seq}[$i])];
	}
	
	my ($CurRefCand, $CurQryCand) = 0;
	my ($rbegin, $rend, $qbegin, $qend, $temp) = (0,0,0,0,0);
	my $nonxbegin = 0;
	
	while(<$DeltaFH>){
        	if(/>(\S+)\s+(\S+)\s+(\d+)\s+(\d+)/){
        			$CurRefCand = $RHRef->{number}{$1};
        			$CurQryCand = $QHRef->{number}{$2};
               		print "procesing the next reference sequence -$1, $2: $CurRefCand,$CurQryCand   \n";
                	next;
        	}
        	if(/(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/){
        		
				$rbegin = $1-1;
				$rend = $2-1;
				$qbegin = $3-1;
				$qend = $4-1;

                # correct orientaion
                if($rbegin > $rend){ $temp = $rbegin; 	$rbegin = $rend;	$rend = $temp; }
                
                if($qbegin > $qend){ $temp = $qbegin; 	$qbegin = $qend;	$qend = $temp; }
                
				for(my $i = $rbegin; $i < $rend; $i++){
					$RefSeqs[$CurRefCand][$i] = 'X';
				}
				
				for(my $i = $qbegin; $i < $qend; $i++){
					$QrySeqs[$CurQryCand][$i] = 'X';
				}	
        	}	
	}
	
	for(my $i = 0; $i < $RHRef->{count}; $i++){
			$nonxbegin = $RHRef->{start}[$i];
			for(my $j = $RHRef->{start}[$i]; $j <= $RHRef->{end}[$i]; $j++ ){
				if($RefSeqs[$i][$j-$RHRef->{start}[$i]] eq 'X'){
					if(($j - $nonxbegin) > $MinIndelSize){
						print $IndelFH "DEL\t".$nonxbegin."\t".($j-1)."\n";
					}
					$nonxbegin = $j+1;
				}
			}
			if(($RHRef->{end}[$i] - $nonxbegin) > $MinIndelSize){
						print $IndelFH "DEL\t".$nonxbegin."\t".($RHRef->{end}[$i])."\n";
			}	
	}	
	
	for(my $i = 0; $i < $QHRef->{count}; $i++){
			$nonxbegin = $QHRef->{start}[$i];
			for(my $j = $QHRef->{start}[$i]; $j <= $QHRef->{end}[$i]; $j++ ){
				if($QrySeqs[$i][$j-$QHRef->{start}[$i]] eq 'X'){
					if(($j - $nonxbegin) > $MinIndelSize){
						print $IndelFH "INS\t".$nonxbegin."\t".($j-1)."\n";
					}
					$nonxbegin = $j+1;
				}
			}
			if(($QHRef->{end}[$i] - $nonxbegin) > $MinIndelSize){
						print $IndelFH "INS\t".$nonxbegin."\t".$QHRef->{end}[$i]."\n";
			}	
	}	
		
}