#! /usr/local/bin/perl

#----------------------------------------------------------------------------------------------
#	Author: Ravi Vijaya Satya, BHSAI
#	File: GetSNPPat.pl
#	Purpose: merges multiple SNP files using the same ref and many strains into one single table file
# 	Also outputs the positions ofeach SNP and its profile, along with the its poisiton in atleast one genome
# 	The input file should have one line for each input strain: 
#	the strain identifier and the strain snp file name, separated by \t
#	the description line of each sequence is expected to contain start: end: length: etc. 
#-----------------------------------------------------------------------------------------------

use strict;
use warnings;
use Bio::Seq;
use Bio::SeqIO;
use File::Temp qw/ tempfile /;

sub read_file_list($);

sub find_if_del($$$);

our (@StrainPolyFileList, @StrainIndelFile, @StrainNameList, @StrainSeq);
our $StrainCnt = 0;
my (%PolyList, %PolyPos, %PolySrc);
my $FH;
my ($PolyType, $RefString, $QString, $RefPos, $QPos, $PolyString, $RefAllele, $QAllele);
my ($RefCode, $index, $dfirst, $dlast);

my (@StartList, @EndList);
if(!@ARGV or $#ARGV < 1){
	die("Usage: perl $0 StrainListFile OutTableFile 
	");
}

# read all the input snp files

our ($InListFile, $OutTableFile) = @ARGV[0..1];

read_file_list($InListFile);

# read the deletion list for each input strain

for(my $i = 0; $i < $StrainCnt; $i++){
	# open the indel file
	open($FH, "<$StrainIndelFile[$i]");
	$index = 0;
	$StartList[$i] 	= [];
	$EndList[$i]	= [];
	while(<$FH>){
		if(/^(\S+)\s+(\d+)\s+(\d+)/){
			($PolyType, $dfirst, $dlast) = ($1, $2, $3);
			if($PolyType eq "DEL"){
				$StartList[$i][$index] = $2;
				$EndList[$i][$index] = $3;
				$index++;
			}
		}
	}
	close($FH);
}

for(my $i = 0; $i < $StrainCnt; $i++){	
	# open the tbl file
	open($FH, "<$StrainPolyFileList[$i]");
	while(<$FH>){
		if(/^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/){
			($PolyType, $RefString, $QString, $RefPos, $QPos, $PolyString) = ($1, $2, $3, $4, $5, $6);
			($RefAllele, $QAllele) = split(/\|/, $PolyString);
			#print $PolyString."\t".$RefAllele."\t".$QAllele."\n";
			$RefCode = $RefPos;
			if($PolyType eq "DEL"){
				$RefCode = $RefPos-0.1;
			}elsif($PolyType eq "INS"){
				$RefCode = $RefPos+0.1;
			}
						
			if(!exists $PolyList{$RefCode}){
				$PolyList{$RefCode} = ();
				$PolyPos{$RefCode} = ();
				$PolySrc{$RefCode} = ();
				for(my $j = 0; $j < $StrainCnt; $j++){
					if(find_if_del($RefPos, $StartList[$j], $EndList[$j]) == -1){
						$PolyList{$RefCode}[$j] = $RefAllele;
					}else{
						$PolyList{$RefCode}[$j] = "X";	# this polymorphic position is entirely missing in the current strain
					}
					$PolyPos{$RefCode}[$j] = 0;			# setting the position to unknown
				}
				$PolyPos{$RefCode}[0] = $RefPos;		# setting the position in the reference
				$PolySrc{$RefCode}[0] = $RefString;	
			}
			$PolyList{$RefCode}[$i] = $QAllele;
			$PolyPos{$RefCode}[$i] = $QPos;
			$PolySrc{$RefCode}[$i] = $QString;
		}
	}
	# close the file
	close ($FH);
}

# read all the sequences into memory
#my $StrainFile;
#for(my $i = 0; $i < $StrainCnt; $i++){
#	$StrainFile = new Bio::SeqIO(-format => 'fasta', -file => $StrainSeqFileList[$i]);
#	$StrainSeq[$i] = ($StrainFile->next_seq())->seq();
#}


# print the snps and polymorphisms in order

my ($OutFH, $OutHybFH, $Pos);

open ($OutFH, ">$OutTableFile");
print $OutFH join ("\t", "Position", "Type", (@StrainNameList))."\n"; 

open($OutHybFH, ">$OutTableFile.hyb");
print $OutHybFH join("\t", "sig_no", "poly_type", "allele", "strain_id", "strain_name", "source", "position", (@StrainNameList))."\n"; 



my $Allele;
my $SigNum = 0;

foreach $RefCode (sort {$a <=> $b} (keys(%PolyList))){
	my %AlleleHybPat;
	my %AlleleData;
	$Pos = int($RefCode +0.5 );			# convert to integer position
	
	if($Pos > $RefCode){				# deletion
		$PolyType = "DEL";
	}elsif($Pos < $RefCode){			# insertion	
		$PolyType = "INS";	
	}else{								# SNP
		$PolyType = "SNP"
	}
	
	print $OutFH $Pos."\t".$PolyType;
	for(my $j = 0; $j < $StrainCnt; $j++){
		print $OutFH "\t".$PolyList{$RefCode}[$j];
		if($PolyList{$RefCode}[$j] ne 'X' && !exists($AlleleHybPat{$PolyList{$RefCode}[$j]})){
			$AlleleHybPat{$PolyList{$RefCode}[$j]} = ();
			for(my $k = 0; $k < $StrainCnt; $k++){ $AlleleHybPat{$PolyList{$RefCode}[$j]}[$k] = 0;}
			$AlleleData{$PolyList{$RefCode}[$j]} = {
				"seq_num" => $j,
				"pos" => $PolyPos{$RefCode}[$j],
				"allele" => $PolyList{$RefCode}[$j],
				"polytype" => $PolyType,
				"source" => $PolySrc{$RefCode}[$j]
			};
		}
		$AlleleHybPat{$PolyList{$RefCode}[$j]}[$j] = 1;
	}
	print $OutFH "\t".$PolySrc{$RefCode}[0]."\n";
	foreach $Allele (keys %AlleleData){
		$SigNum++;
		print $OutHybFH join("\t", $SigNum, $AlleleData{$Allele}->{polytype}, $AlleleData{$Allele}->{allele}, 
							1+$AlleleData{$Allele}->{seq_num}, $StrainNameList[$AlleleData{$Allele}->{seq_num}], 
							$AlleleData{$Allele}->{source}, $AlleleData{$Allele}->{pos}, @{$AlleleHybPat{$Allele}})."\n"; 
	}
	
}

close $OutFH;
close $OutHybFH;

# write the pairwise distance matrix
my (@DistanceMtx1, @DistanceMtx2, @DistanceMtx3);

for(my $i = 0; $i < $StrainCnt; $i++){
	$DistanceMtx1[$i] = ();
	$DistanceMtx2[$i] = ();
	$DistanceMtx3[$i] = ();
	for(my $j = 0; $j < $StrainCnt ; $j++){
		$DistanceMtx1[$i][$j] = 0;
		$DistanceMtx2[$i][$j] = 0;
		$DistanceMtx3[$i][$j] = 0;
	}
}

foreach $RefCode (keys (%PolyList)){
	$Pos = int($RefCode+0.5);
	my $NoX = 0;
	for(my $i = 1; $i < $StrainCnt; $i++){
		for(my $j = 0; $j < $i ; $j++){
			if($PolyList{$RefCode}[$i] ne $PolyList{$RefCode}[$j]){
				$DistanceMtx1[$i][$j]++;
				$DistanceMtx1[$j][$i]++;
				$DistanceMtx2[$i][$j]++;
				$DistanceMtx2[$j][$i]++
			}elsif($PolyList{$RefCode}[$i] eq 'X'){	# if both of them are 'X', count it as a mismatch
				$DistanceMtx2[$i][$j]++;
				$DistanceMtx2[$j][$i]++;
			}
			if($PolyList{$RefCode}[$j] eq 'X'){
				$NoX++;
			}
		}
	}
	if($NoX == 0){
		for(my $i = 1; $i < $StrainCnt; $i++){
			for(my $j = 0; $j < $i ; $j++){
				if($PolyList{$RefCode}[$i] ne $PolyList{$RefCode}[$j]){
					$DistanceMtx3[$i][$j]++;
					$DistanceMtx3[$j][$i]++;
				}
			}
		}	
	}
}

open($OutFH, ">".$OutTableFile.".xmatch.dmtx");
for(my $i = 0; $i < $StrainCnt; $i++){
	print $OutFH "\t".$StrainNameList[$i];
}
print $OutFH "\n";
my $Min = 0;
for(my $i = 0; $i < $StrainCnt; $i++){
	$Min=0;
	if($i == 0) {$Min = 1};
	print $OutFH $StrainNameList[$i];
	for(my $j = 0; $j < $StrainCnt ; $j++){
		print $OutFH "\t".$DistanceMtx1[$i][$j];
		if($i != $j && $DistanceMtx1[$i][$j] < $DistanceMtx1[$i][$Min]){
			$Min = $j;
		}		
	}
	print $OutFH "\t".$StrainNameList[$Min]."\n";
}

close $OutFH;



open($OutFH, ">".$OutTableFile.".xmiss.dmtx");
for(my $i = 0; $i < $StrainCnt; $i++){
	print $OutFH "\t".$StrainNameList[$i];
}
print $OutFH "\n";
for(my $i = 0; $i < $StrainCnt; $i++){
	$Min=0;
	if($i == 0) {$Min = 1};
	print $OutFH $StrainNameList[$i];
	for(my $j = 0; $j < $StrainCnt ; $j++){
		print $OutFH "\t".$DistanceMtx2[$i][$j];
		if($i != $j && $DistanceMtx2[$i][$j] < $DistanceMtx2[$i][$Min]){
			$Min = $j;
		}		
	}
	print $OutFH "\t".$StrainNameList[$Min]."\n";
}

close $OutFH;


open($OutFH, ">".$OutTableFile.".dmtx");
for(my $i = 0; $i < $StrainCnt; $i++){
	print $OutFH "\t".$StrainNameList[$i];
}
print $OutFH "\tClosest\n";
for(my $i = 0; $i < $StrainCnt; $i++){
	$Min=0;
	if($i == 0) {$Min = 1};
	print $OutFH $StrainNameList[$i];
	for(my $j = 0; $j < $StrainCnt ; $j++){
		print $OutFH "\t".$DistanceMtx3[$i][$j];
		if($i != $j && $DistanceMtx3[$i][$j] < $DistanceMtx3[$i][$Min]){
			$Min = $j;
		}		
	}
	print $OutFH "\t".$StrainNameList[$Min]."\n";
}

close $OutFH;

sub read_file_list($){
	my $InFileName = $_[0];

	my $INFH;
	
	open($INFH, "<$InFileName");
	$StrainCnt = 0;
	while(<$INFH>){
		if(/(\S+)\s+(\S+)\s+(\S+)/){
			$StrainPolyFileList[$StrainCnt] = $1;
			$StrainIndelFile[$StrainCnt] = $2;
			$StrainNameList[$StrainCnt] = $3;
			chomp($StrainNameList[$StrainCnt]);
			$StrainCnt++;
		}
	} 	
	close($INFH);
}

sub find_if_del($$$){
	#my ($Value, @Starts, @Ends)  = ($_[0], @{$_[1]},  ${$_[2]});
	my ($Value, $StartRef, $EndRef)  = ($_[0], $_[1],  $_[2]);
	my @Starts 	= @{$StartRef};
	my @Ends 	= @{$EndRef};
	my $MinPos = 0;
	my $MaxPos = $#Starts;
	
	if($MaxPos < 0 ){
		return -1;		# not in the array
	}
	
	if($Starts[$MinPos] > $Value) {
		return -1; 		# not in the array
	}
	
	if($Value > $Ends[$MaxPos]){
		return -1;
	}
	
	while(1){		
		my $mid = int(($MinPos + $MaxPos)/2);
		
		if($Value >= $Starts[$mid]){
			if($Value <= $Ends[$mid]){
				return 1;		# in the range
			}else{
				$MinPos = $mid+1;
				if($MinPos > $MaxPos){
					return -1;
				} 
			}
		}else{
			$MaxPos = $mid -1;
			if($MaxPos < $MinPos){
				return -1;
			}
		}
	}
}
