#! /usr/local/bin/perl

# This script reads in (1) a sequence file containing reads, contigs or scaffolds in FASTA format, and:
#  					   (2) a file with quality scores in phred format
# The output will be a sequence file in FASTA format with all the bases with low quality scores replaced
# by 'N's

use strict;
use Bio::Seq;
use Bio::SeqIO;

if(!@ARGV or $#ARGV < 2){
	die("Usage: $0 input_seq_file input_qual_file output seq_file_quality_threshold[optional]");
}

my ($InSeqFileName, $InQualFileName, $OutSeqFileName) = @ARGV[0..2];
my $QualThreshold = 20;
if($#ARGV >=3){
	if($ARGV[3] >=0 && $ARGV[3] <= 93){
		$QualThreshold = $ARGV[3];
	}else{
		print("Input quality threshold $ARGV[3] is outside the range 0-93. Switching to the default threshold of 20\n");
	}
}

my ($InSeqFile, $InQualFile, $OutSeqFile);

if(!(-e $InSeqFileName)){
	die("The input file  $InSeqFileName does not exist");
}

if(!(-e $InQualFileName)){
	die("The input file $InQualFileName does not exist");
}

eval{ 
	$InSeqFile = new Bio::SeqIO(-format => 'fasta', -file => $InSeqFileName);
};

if($@){
	die("Problem opening fasta format file $InSeqFileName", $@);
}

eval{ 
	$InQualFile = new Bio::SeqIO(-format => 'qual', -file => $InQualFileName);
};

if($@){
	die("Problem opening phred format file $InQualFileName", $@);
}

eval{ 
	$OutSeqFile = new Bio::SeqIO(-format => 'fasta', -file => ">$OutSeqFileName");
};

if($@){
	die("Problem opening $OutSeqFileName for writing", $@);
}


my (@ContigHdrs, @ContigSizes);
my %ContigSeqs;
my ($ContigCnt, $TotalSize) = (0,0);
while(my $NextSeq = $InSeqFile->next_seq()){
	$ContigHdrs[$ContigCnt] = $NextSeq->id();
	$ContigSizes[$ContigCnt] = $NextSeq->length();
	$ContigSeqs{$NextSeq->id()} = $NextSeq->seq();
	$TotalSize += $NextSeq->length();
	$ContigCnt++;
}
my  (@ContigChars, @ContigQuals);
my $NewContigSeq;
while(my $NextSeq = $InQualFile->next_seq()){
	@ContigChars = split(//, $ContigSeqs{$NextSeq->id()});
	@ContigQuals = @{$NextSeq->qual()};
	
	for(my $i = 0; $i < $NextSeq->length(); $i++){
		if($ContigQuals[$i] < $QualThreshold){
			#$ContigChars[$i] = 'N';
			$ContigChars[$i] = lc($ContigChars[$i]);			
		}
	}
	$NewContigSeq = Bio::Seq->new(-format => 'fasta', -id => $NextSeq->id(), -desc => $NextSeq->desc(), -seq => join('', @ContigChars));
	$OutSeqFile->write_seq($NewContigSeq);
}
