#! /usr/local/bin/perl

#############################################################################
#	Author: Ravi Vijaya Satya
# 	This perl script takes a list of input files and performs SNP discovery on them
##############################################################################

use Cwd;
use Env;
use strict;
use warnings;
use Getopt::Std;

my $SourceDirectory=$SNIT_HOME."/src";

my $DEF_mask_tandem_repeats     = 0;
my $DEF_trf_flags		= "2 5 5 80 10 50 500 -h";
my $DEF_mmr_cluster_len         = 201;
my $DEF_mmr_min_exact_match     = 100;
my $DEF_mmr_max_gap             = 1;
my $DEF_min_indel_size          = 100;
my $DEF_min_conserved_flank_len = 100;
my $DEF_min_dist_from_edges     = 250;
my $DEF_ignore_lc_qual          = 0;     

sub usage;
sub get_user_input;
sub load_target_genome_list($);

#------------------------------------------------------------------------------------------------
if($#ARGV < 1) { usage();}

my ($config_file, $out_prefix, $work_dir, $result_dir, $mask_tandem_repeats, $trf_flags, $mmr_cluster_len, $mmr_min_exact_match, 
	$mmr_max_gap, $min_indel_size, $min_conserved_flank_len, $min_dist_from_edges, $ignore_lc_qual) = get_user_input(@ARGV);

#-------------------------------------------------------------------------------------------------

my @TargetGenomeFileList;
my @TargetGenomeIdentifierList;
my $NoOfTargets = 0;

print "trying to cd $work_dir\n";

#change to work directory
chdir($work_dir) or die "$!";

system("date");

#load the input genome list file
&load_target_genome_list($config_file);

my $GenomePrefix;
# prepare each input file
for(my $i = 0; $i < $NoOfTargets; $i++){
	$GenomePrefix = sprintf("g%03d",($i+1) );
	system("perl $SourceDirectory/PrepareMummerInput.pl $TargetGenomeFileList[$i] $GenomePrefix $TargetGenomeIdentifierList[$i].cand");
}

if($mask_tandem_repeats == 1){
	for(my $i = 0; $i < $NoOfTargets; $i++){
		system("$SNIT_HOME/bin/trf.exe $TargetGenomeIdentifierList[$i].cand ".$trf_flags);
		system("mv $TargetGenomeIdentifierList[$i].cand.*.dat $TargetGenomeIdentifierList[$i].trf.dat");
		system("perl $SourceDirectory/MaskTRepeats.pl $TargetGenomeIdentifierList[$i].cand $TargetGenomeIdentifierList[$i].trf.dat $TargetGenomeIdentifierList[$i].trf.out");
		system("mv $TargetGenomeIdentifierList[$i].trf.out $TargetGenomeIdentifierList[$i].cand");
	}
}


# pick genome 0 as the reference
my $RefSeq = $TargetGenomeIdentifierList[0].".cand";

for(my $i = 1; $i < $NoOfTargets; $i++){
	system("perl $SourceDirectory/FindSNPS1bpIndelsGlobalPos.pl $RefSeq $TargetGenomeIdentifierList[$i].cand $mmr_cluster_len $mmr_min_exact_match $mmr_max_gap $min_indel_size $TargetGenomeIdentifierList[0]_$TargetGenomeIdentifierList[$i]");
	system("perl $SourceDirectory/FilterSNPsGlobalPos.pl $TargetGenomeIdentifierList[0]_$TargetGenomeIdentifierList[$i].poly $min_conserved_flank_len $min_dist_from_edges  $ignore_lc_qual $TargetGenomeIdentifierList[0]_$TargetGenomeIdentifierList[$i].poly.filtered");
}

# create dummy files for ref genome
system("echo -n \" \" >> $TargetGenomeIdentifierList[0]_$TargetGenomeIdentifierList[0].poly.filtered");
system("echo -n \" \" >> $TargetGenomeIdentifierList[0]_$TargetGenomeIdentifierList[0].lindels");


my ($OutConfigFH, $CandListFH);
open($OutConfigFH, ">$out_prefix.slist");
open($CandListFH, ">$out_prefix.cand.list");

for(my $i = 0; $i < $NoOfTargets; $i++){
	print $OutConfigFH $TargetGenomeIdentifierList[0]."_".$TargetGenomeIdentifierList[$i].".poly.filtered\t".$TargetGenomeIdentifierList[0]."_".$TargetGenomeIdentifierList[$i].".lindels\t".$TargetGenomeIdentifierList[$i]."\n";
	print $CandListFH "$TargetGenomeIdentifierList[$i].cand\t".$TargetGenomeIdentifierList[$i]."\n";
}

close($OutConfigFH);
close($CandListFH);

#system("perl $SourceDirectory/GetSNPPat.pl $out_prefix.slist $out_prefix.tbl");
system("java snit.ConstructSNPTable $out_prefix.slist $out_prefix.tbl $NoOfTargets");
system("perl $SourceDirectory/ConvertSNPIndices.pl $TargetGenomeIdentifierList[0].cand $out_prefix.tbl $out_prefix.tbl.final");

print "--------------------------------------------------------\n";
system("cp $work_dir/$out_prefix.tbl.dmtx $result_dir/$out_prefix.tbl.dmtx");
system("cp $work_dir/$out_prefix.tbl.final $result_dir/$out_prefix.tbl.final");
system("cp $work_dir/$out_prefix.tbl.xmatch.dmtx $result_dir/$out_prefix.tbl.xmatch.dmtx");
system("cp $work_dir/$out_prefix.tbl.xmiss.dmtx $result_dir/$out_prefix.tbl.xmiss.dmtx");

print "Successfully completed all steps\n";
print "The file $result_dir/$out_prefix.tbl.dmtx contains the numbers of discriminating SNPs between each pair of genomes\n";
print "The file  $result_dir/$out_prefix.tbl.final contains the detailed list of SNPs\n";
system("cat $result_dir/$out_prefix.tbl.dmtx");
system("date");


#############################################################################
#### Load the target genome list
#############################################################################
sub load_target_genome_list($){
	my $ListFile = $_[0];
	$NoOfTargets = 0;
	my ($TargetFile, $TargetName);
	my $ErrorFlag = 0;
	open(LISTFILE, $ListFile) or die "could not open $ListFile";
	while(<LISTFILE>){
		if(/^FILE=(.+), NAME=(.*)/){
			$TargetFile = $1;
			$TargetName = lc($2);
			if(!(-e $TargetFile)){
				print "The target file $TargetFile does not exist\n";
				print "Please make sure that the file exists and re-run the program\n";
				$ErrorFlag = 1;				
			}
			#if(!($TargetName =~ /(\S+)/)){
			#	print "The target genome identifier for file $TargetFile must have atleast one phrase\n";
			#	print "Please re-run with a proper name for each file\n";
			#	$ErrorFlag=1;
			#}
			$TargetGenomeIdentifierList[$NoOfTargets] = $TargetName;
			$TargetGenomeFileList[$NoOfTargets] = $TargetFile;
			$NoOfTargets++;
		}
	}
	
	if($ErrorFlag == 1){
		die("There were errors reading the target genome list file");
	}
	
	#DEBUG
	print "Target Genomes: \n";
	for(my $i = 0; $i < $NoOfTargets; $i++){
		print "File: $TargetGenomeFileList[$i], Identifier: $TargetGenomeIdentifierList[$i]\n";
	}
}

####################################################################################
#### print usage
####################################################################################
sub usage(){
die <<EOF

USAGE
SNIT.pl -i <config file> -o <output prefix>
OPTIONS
        -h                   print usage
        -i    string         configuration file name
        -o    string         output prefix
	-w    string 	     output (work) directory
	-r    string	     output (result directory)
        -f    [0/1]          Screen input files with trf 0- no, 1 yes [default: $DEF_mask_tandem_repeats]
        -t    string         trf flags default: "$DEF_trf_flags" (ignored if run with -f 0)
        -c    int            minimum MUMmer cluster length [default: $DEF_mmr_cluster_len ]
        -m    int            minimum MUMmer exact match length [default: $DEF_mmr_min_exact_match]
        -g    int            maximum MUMmer gap length for extension of alignments [default: $DEF_mmr_max_gap]
        -d    int            min indel size to report as insertion/deletion [default: $DEF_min_indel_size]
        -s    int            min conserved flank length on either side of a SNP [default: $DEF_min_conserved_flank_len]
        -e    int            min distance of a SNP from the edges of a contig [default: $DEF_min_dist_from_edges]
        -q    [0/1]          interpret smaller case bases as low quality bases and ignore SNPs in these bases 0-no, 1-yes [default: $DEF_ignore_lc_qual]
EOF
}

#####################################################################################
#### Get user input using getopt
#####################################################################################

sub get_user_input{
	my %options;
	getopts('h:i:o:w:r:f:t:c:m:g:d:s:e:q:', \%options);
	
	my ($config_file, $out_prefix, $work_dir, $result_dir, $mask_tandem_repeats, $trf_flags, $mmr_cluster_len, $mmr_min_exact_match, 
	$mmr_max_gap, $min_indel_size, $min_conserved_flank_len, $min_dist_from_edges, $ignore_lc_qual);
	
	if(exists $options{h}){
		usage();
		exit(0);
	}
	if((exists $options{i}) && (exists $options{o}) ) {
		($config_file, $out_prefix) = ($options{i}, $options{o});
	}else{
		usage();
	}
	
	if(exists $options{w}){ $work_dir = $options{w};} else {$work_dir = "."; }
	
	if(exists $options{r}){ $result_dir = $options{r};} else {$result_dir = "."; }

	if(exists $options{f}){ $mask_tandem_repeats = $options{f};} else {$mask_tandem_repeats = $DEF_mask_tandem_repeats; }
	
	if(exists $options{t}){ $trf_flags = $options{t}; } else { $trf_flags = $DEF_trf_flags; }
	
	if(exists $options{c}){ $mmr_cluster_len = $options{c}; } else { $mmr_cluster_len = $DEF_mmr_cluster_len; }
	
	if(exists $options{m}){ $mmr_min_exact_match = $options{m}; } else { $mmr_min_exact_match = $DEF_mmr_min_exact_match; }
	
	if(exists $options{g}){ $mmr_max_gap = $options{g}; } else { $mmr_max_gap = $DEF_mmr_max_gap; }
	
	if(exists $options{d}){ $min_indel_size = $options{d}; } else { $min_indel_size = $DEF_min_indel_size; }
	
	if(exists $options{s}){ $min_conserved_flank_len = $options{s}; } else { $min_conserved_flank_len = $DEF_min_conserved_flank_len; }
	
	if(exists $options{e}){ $min_dist_from_edges = $options{e}; } else { $min_dist_from_edges = $DEF_min_dist_from_edges; }
	
	if(exists $options{q}){ $ignore_lc_qual = $options{q}; } else {$ignore_lc_qual = $DEF_ignore_lc_qual; }
	
	return ($config_file, $out_prefix, $work_dir, $result_dir, $mask_tandem_repeats, $trf_flags, $mmr_cluster_len, $mmr_min_exact_match, 
	$mmr_max_gap, $min_indel_size, $min_conserved_flank_len, $min_dist_from_edges, $ignore_lc_qual);
	
}
